/*
 * Copyright (C) 2008 by egnite GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Id$
 */

#include <stdlib.h>
#include <string.h>

#include "eirhwt.h"

#define DF_PAGE_SIZE        528
#define DF_PAGES            8192

#define DF_WRITE_POLLS        1000

#define DFCMD_READ_STATUS       0xD7
#define DFCMD_CONT_READ         0xE8
#define DFCMD_BUF1_WRITE        0x84
#define DFCMD_BUF1_FLASH        0x83
#define DFCMD_PAGE_ERASE        0x81

int SpiInit(void)
{
    /* Enable peripherals. */
    outr(PIOA_ASR, inr(PIOA_ASR) 
        | _BV(PA12_SPI0_MISO_A) 
        | _BV(PA13_SPI0_MOSI_A)
        | _BV(PA14_SPI0_SPCK_A)
        //| _BV(PA31_SPI0_NPCS1_A);
        | _BV(PA30_IRQ1_A));

    /* Disable PIO lines used for SPI. */
    outr(PIOA_PDR, _BV(PA12_SPI0_MISO_A)
        | _BV(PA13_SPI0_MOSI_A)
        | _BV(PA14_SPI0_SPCK_A)
        //| _BV(PA31_SPI0_NPCS1_A);
        | _BV(PA30_IRQ1_A));

    /* Enable SPI clock. */
    outr(PMC_PCER, _BV(SPI0_ID) | _BV(PIOA_ID) | _BV(PIOB_ID) | _BV(PIOC_ID));

    /* SPI enable and reset. */
    outr(SPI0_CR, SPI_SPIEN | SPI_SWRST);
    outr(SPI0_CR, SPI_SPIEN);

    outr(PIOA_PER, _BV(PA11_SPI0_NPCS0_A));
    outr(PIOA_SODR, _BV(PA11_SPI0_NPCS0_A));
    outr(PIOA_OER, _BV(PA11_SPI0_NPCS0_A));

    return 0;
}

int SpiSetMode(void)
{
    /* Set SPI to master mode, fixed peripheral, fault detection disabled. */
    outr(SPI0_MR, SPI_MODFDIS | SPI_MSTR);

    /* Data changes during clock low and will be sampled on rising edges. */
    outr(SPI0_CSR0, (50 << SPI_SCBR_LSB) | SPI_CPOL);

    return 0;
}

static unsigned char SpiByte(unsigned char val)
{
    /* Transmission is started by writing the transmit data. */
    outr(SPI0_TDR, val);
    /* Wait for receiver data register full. */
    while((inr(SPI0_SR) & SPI_RDRF) == 0);
    /* Read data. */
    val = (unsigned char)inr(SPI0_RDR);

    return val;
}

int SpiTransfer2(CONST void *txbuf, void *rxbuf, int xlen, CONST void *txnbuf, void *rxnbuf, int xnlen)
{
    unsigned char *txp;
    unsigned char *rxp;
    unsigned char b;

    outr(PIOA_CODR, _BV(PA11_SPI0_NPCS0_A));
    txp = (unsigned char *)txbuf;
    rxp = (unsigned char *)rxbuf;
    b = 0xff;
    while (xlen--) {
        if (txp) {
            b = *txp++;
        }
        b = SpiByte(b);
        if (rxp) {
            *rxp++ = b;
        }
    }
    txp = (unsigned char *)txnbuf;
    rxp = (unsigned char *)rxnbuf;
    b = 0xff;
    while (xnlen--) {
        if (txp) {
            b = *txp++;
        }
        b = SpiByte(b);
        if (rxp) {
            *rxp++ = b;
        }
    }
    outr(PIOA_SODR, _BV(PA11_SPI0_NPCS0_A));

    return 0;
}

unsigned char DataFlashGetStatus(void)
{
    unsigned char buf[2] = { DFCMD_READ_STATUS, 0xFF };

    if (SpiTransfer2(buf, buf, 2, NULL, NULL, 0)) {
        return (unsigned char) - 1;
    }
    return buf[1];
}

#define MAX_DF_CMDLEN         8

static unsigned char dcb_cmdbuf[MAX_DF_CMDLEN];

int DataFlashSendCmd(unsigned char op, unsigned long parm, int len, CONST void *tdata, void *rdata, int datalen)
{
    unsigned char *cb = dcb_cmdbuf;

    if (len > MAX_DF_CMDLEN) {
        return -1;
    }
    memset(cb, 0, len);
    cb[0] = op;
    if (parm) {
        cb[1] = (unsigned char) (parm >> 16);
        cb[2] = (unsigned char) (parm >> 8);
        cb[3] = (unsigned char) parm;
    }
    return SpiTransfer2(cb, cb, len, tdata, rdata, datalen);
}

int DataFlashWaitReady(unsigned long tmo, int poll)
{
    unsigned char sr;

    while (((sr = DataFlashGetStatus()) & 0x80) == 0) {
        if (!poll) {
            Delay(1);
        }
        if (tmo-- == 0) {
            return -1;
        }
    }
    return 0;
}

static unsigned char pgbuf[DF_PAGE_SIZE];
static unsigned char rpbuf[DF_PAGE_SIZE];

int DataFlashPageRead(unsigned long pgn, void *data, unsigned int len)
{
    pgn <<= 10;
    return DataFlashSendCmd(DFCMD_CONT_READ, pgn, 8, data, data, len);
}

int DataFlashPageWrite(unsigned int pgn, CONST void *data, unsigned int len)
{
    int rc = -1;

    /* Copy data to dataflash RAM buffer. */
    if (DataFlashSendCmd(DFCMD_BUF1_WRITE, 0, 4, data, rpbuf, len) == 0) {
        /* Flash RAM buffer. */
        pgn <<= 10;
        if (DataFlashSendCmd(DFCMD_BUF1_FLASH, pgn, 4, NULL, NULL, 0) == 0) {
            rc = DataFlashWaitReady(DF_WRITE_POLLS, 1);
        }
    }
    return rc;
}

int DataFlashPageErase(unsigned int pgn)
{
    pgn <<= 10;
    return DataFlashSendCmd(DFCMD_PAGE_ERASE, pgn, 4, NULL, NULL, 0);
}

int TestDataFlash(void)
{
    unsigned long pgn;
    int k;

    SpiInit();
    SpiSetMode();

    k = 0;
    printf("Testing Dataflash");
    for (pgn = 0; pgn < DF_PAGES; pgn++) {
        if ((pgn & 0x7f) == 0) {
            putchar('.');
        }
        DataFlashPageErase(pgn);
        memset(pgbuf, 0, sizeof(pgbuf));
        DataFlashPageRead(pgn, pgbuf, sizeof(pgbuf));
        if (pgbuf[0] != 0xff) {
            printf("Erases %lu to %02x\n", pgn, pgbuf[0]);
            return -1;
        }

        memset(pgbuf, k, sizeof(pgbuf));
        DataFlashPageWrite(pgn, pgbuf, sizeof(pgbuf));
        memset(pgbuf, 0xff, sizeof(pgbuf));
        DataFlashPageRead(pgn, pgbuf, sizeof(pgbuf));
        if (pgbuf[0] != (unsigned char)k) {
            puts("Error");
            printf("Writing page %lu failed\n", pgn);
            return -1;
        }
        k++;
#if 0
        for (i = 0; i < sizeof(pgbuf); i++) {
            printf("%02x", pgbuf[i]);
        }
        putchar('\n');
#endif
    }
    puts("OK");
#if 0
    for (pgn = 0; pgn < DF_PAGES; pgn++) {
        DataFlashPageRead(pgn, pgbuf, sizeof(pgbuf));
        if ((pgn & 31) == 0) {
            putchar('\n');
        }
        printf("%02x ", pgbuf[0]);
    }
#endif
    return 0;
}
