/*
 * Copyright (C) 2008 by egnite GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Id$
 */

#include <stdlib.h>
#include <string.h>

#include "eirhwt.h"

#ifndef SPI_CS_BIT
/*! \brief Port bit for SPI chip select. */
#define SPI_CS_BIT     15
#endif

#ifndef SPI_CLK_BIT
/*! \brief Port bit for SPI clock. */
#define SPI_CLK_BIT    16
#endif

#ifndef SPI_MOSI_BIT
/*! \brief Port bit for SPI MOSI line. */
#define SPI_MOSI_BIT   17
#endif

#ifndef SPI_MISO_BIT
/*! \brief Port bit for SPI MISO line. */
#define SPI_MISO_BIT   18
#endif

#ifndef SPI_PE_REG
/*! \brief SPI port enable register.
 */
#define SPI_PE_REG     PIOA_PER
#endif

#ifndef SPI_PD_REG
/*! \brief SPI port disable register.
 */
#define SPI_PD_REG     PIOA_PDR
#endif

#ifndef SPI_OE_REG
/*! \brief SPI output enable register. */
#define SPI_OE_REG     PIOA_OER
#endif

#ifndef SPI_OD_REG
/*! \brief SPI output disable register. */
#define SPI_OD_REG     PIOA_ODR
#endif

#ifndef SPI_SOD_REG
/*! \brief SPI set output data register. */
#define SPI_SOD_REG    PIOA_SODR
#endif

#ifndef SPI_COD_REG
/*! \brief SPI clear output data register. */
#define SPI_COD_REG    PIOA_CODR
#endif

#ifndef SPI_PDS_REG
/*! \brief SPI port data status register. */
#define SPI_PDS_REG    PIOA_PDSR
#endif

#ifndef SPI_ODS_REG
/*! \brief SPI port output status register. */
#define SPI_ODS_REG    PIOA_ODSR
#endif


#define MMCMD_HOST                      0x40
#define MMCMD_RESET_CRC                 0x95

/*! \brief Reset card to idle state. 
 *
 * In idle state the card will not accept any other commands than
 * MMCMD_SEND_OP_COND or MMCMD_READ_OCR.
 */
#define MMCMD_GO_IDLE_STATE             0

/*! \brief Activate card's initialization process. */
#define MMCMD_SEND_OP_COND              1

/*! \brief Send operation condition register.
 *
 * Queries the operation condition register content from all cards, which
 * are in idle state.
 */
#define MMCMD_ALL_SEND_CID              2

/*! \brief Assign relative card address. */
#define MMCMD_SEND_RELATIVE_ADDR        3

/*! \brief Assign relative card address. */
#define MMCMD_SELECT_CARD               7

/*! \brief Query card's CSD. */
#define MMCMD_SEND_CSD                  9

/*! \brief Query card's CID. */
#define MMCMD_SEND_CID                  10

/*! \brief Stop multiple block transmission. */
#define MMCMD_STOP_TRANSMISSION         12

/*! \brief Query card's status register. */
#define MMCMD_SEND_STATUS               13

/*! \brief Select block length for following read/write commands. */
#define MMCMD_SET_BLOCKLEN              16

/*! \brief Initiate single block read. */
#define MMCMD_READ_SINGLE_BLOCK         17

/*! \brief Initiate continuous block read. */
#define MMCMD_READ_MULTIPLE_BLOCK       18

/*! \brief Initiate single block write. */
#define MMCMD_WRITE_BLOCK               24

/*! \brief Initiate continuous block write. */
#define MMCMD_WRITE_MULTIPLE_BLOCK      25

/*! \brief Initiate programming of programmable CSD bits. */
#define MMCMD_PROGRAM_CSD               27

/*! \brief Enable card's optional write protection. */
#define MMCMD_SET_WRITE_PROTECT         28

/*! \brief Disable card's write protection. */
#define MMCMD_CLR_WRITE_PROTECT         29

/*! \brief Query card's write protect status. */
#define MMCMD_SEND_WRITE_PROTECT        30

/*! \brief Set address of the first erase group. */
#define MMCMD_TAG_ERASE_GROUP_START     35

/*! \brief Set address of the last erase group. */
#define MMCMD_TAG_ERASE_GROUP_END       36

/*! \brief Erase previously selected sectors. */
#define MMCMD_ERASE                     38

/*! \brief Set/clear password or lock/unlock the card. */
#define MMCMD_LOCK_UNLOCK               42

/*! \brief Query card's operating condition register. */
#define MMCMD_READ_OCR                  58

/*! \brief Enable or disable CRC mode. 
 *
 * In SPI mode CRC is disabled by default.
 */
#define MMCMD_CRC_ON_OFF                59

/*! \brief Card is idle. */
#define MMR1_IDLE_STATE         0x00
/*! \brief Card is busy. */
#define MMR1_NOT_IDLE           0x01
/*! \brief Erase sequence was cleared before execution. */
#define MMR1_ERASE_RESET        0x02
/*! \brief Illegal command code detected. */
#define MMR1_ILLEGAL_COMMAND    0x04
/*! \brief Bad command CRC detected. */
#define MMR1_COM_CRC_ERROR      0x08
/*! \brief Bad erase sequence. */
#define MMR1_ERASE_SEQ_ERROR    0x10
/*! \brief Misaligned address did not match block length. */
#define MMR1_ADDRESS_ERROR      0x20
/*! \brief Command parameter is out of range. */
#define MMR1_PARAMETER_ERROR    0x40


/*! \brief Card is locked. */
#define MMR2_CARD_LOCKED        0x01
/*! \brief Erasing write protected sector or password error. */
#define MMR2_WP_ERASE_SKIP      0x02
/*! \brief General or unknown error occured. */
#define MMR2_ERROR              0x04
/*! \brief Internal card controller error. */
#define MMR2_CC_ERROR           0x08
/*! \brief Bad internal ECC. */
#define MMR2_ECC_FAILED         0x10
/*! \brief Failed to write to protected block. */
#define MMR2_WP_VIOLATION       0x20
/*! \brief Invalid erase parameter. */
#define MMR2_ERASE_PARAMETER    0x40
/*! \brief Command parameter is out of range. */
#define MMR2_OUT_OF_RANGE       0x80


#define MMCSR_OUT_OF_RANGE      0x80000000
#define MMCSR_ADDRESS_ERROR     0x40000000
#define MMCSR_BLOCK_LEN_ERROR   0x20000000
#define MMCSR_ERASE_SEQ_ERROR   0x10000000
#define MMCSR_ERASE_PARAM       0x08000000
#define MMCSR_WP_VIOLATION      0x04000000
#define MMCSR_COM_CRC_ERROR     0x00800000
#define MMCSR_ILLEGAL_COMMAND   0x00400000
#define MMCSR_ERROR             0x00080000
#define MMCSR_CIDCSD_OVERWRITE  0x00010000
#define MMCSR_WP_ERASE_SKIP     0x00008000
#define MMCSR_CARD_ECC_DISABLED 0x00004000
#define MMCSR_ERASE_RESET       0x00002000
#define MMCSR_STATE_MASK        0x00001E00
#define MMCSR_READY_FOR_DATA    0x00000100

#define MMCSR_IS_IDLE           0x00000000
#define MMCSR_IS_READY          0x00000200
#define MMCSR_IS_IDENT          0x00000400
#define MMCSR_IS_STBY           0x00000600
#define MMCSR_IS_TRAN           0x00000800
#define MMCSR_IS_DATA           0x00000A00
#define MMCSR_IS_RCV            0x00000C00
#define MMCSR_IS_PRG            0x00000E00
#define MMCSR_IS_DIS            0x00001000

#define MMDR_ACCEPTED
#define MMDR_CRC_ERROR
#define MMDR_WRITE_ERROR

#define MMCERR_TIMEOUT      0x00000001

#define MMCOP_NBUSY         0x80000000

/*! \brief Number of bytes in the CID register. */
#define MMCARD_CIDR_SIZE                16

/*! \brief Number of bytes in the CSD register. */
#define MMCARD_CSDR_SIZE                16

/*! \brief Number of bytes in the operating condition register. */
#define MMCARD_OCR_SIZE                 32

/*!
 * \brief Multimedia card identification register.
 */
typedef struct __attribute__ ((packed)) _MMC_CID {
    /*! \brief Manufacturer identifier. */
    unsigned char mmcid_mid;
    /*! \brief OEM/Application identifier. */
    unsigned short mmcid_oid;
    /*! \brief Product name. */
    unsigned char mmcid_pnm[6];
    /*! \brief Product revision. */
    unsigned char mmcid_rev;
    /*! \brief Serial number. */
    unsigned long mmcid_psn;
    /*! \brief Manufacturing date code. */
    unsigned char mmcid_mdt;
    /*! \brief CRC7 checksum. */
    unsigned char mmcid_crc;
} MMC_CID;



/*! \brief Set specified MMC line low. */
#define MMC0_CLR_BIT(n)  outr(SPI_COD_REG, _BV(n))
/*! \brief Set specified MMC line high. */
#define MMC0_SET_BIT(n)  outr(SPI_SOD_REG, _BV(n))

/*! \brief Query status of a specified MMC output. */
#define MMC0_IS_BIT_SET(n)   ((inr(SPI_ODS_REG) & _BV(n)) == _BV(n))

/*! \brief Query status of a specified MMC input. */
#define MMC0_TST_BIT(n)  ((inr(SPI_PDS_REG) & _BV(n)) == _BV(n))

#define SPI_RBIT(o, n) \
    outr(SPI_SOD_REG, _BV(SPI_CLK_BIT)); \
    if (inr(PIOA_PDSR) & _BV(SPI_MISO_BIT)) { \
        o |= _BV(n); \
    } \
    outr(SPI_COD_REG, _BV(SPI_CLK_BIT));

#define SPI_RBIT_FIRST(o, n) \
    outr(SPI_SOD_REG, _BV(SPI_MOSI_BIT)); \
    SPI_RBIT(o, n);

#define SPI_R8BIT(o) \
    SPI_RBIT_FIRST(o, 7); \
    SPI_RBIT(o, 6); \
    SPI_RBIT(o, 5); \
    SPI_RBIT(o, 4); \
    SPI_RBIT(o, 3); \
    SPI_RBIT(o, 2); \
    SPI_RBIT(o, 1); \
    SPI_RBIT(o, 0);

#define SPI_R32BIT(o) \
    SPI_RBIT_FIRST(o, 7); \
    SPI_RBIT(o, 6); \
    SPI_RBIT(o, 5); \
    SPI_RBIT(o, 4); \
    SPI_RBIT(o, 3); \
    SPI_RBIT(o, 2); \
    SPI_RBIT(o, 1); \
    SPI_RBIT(o, 0); \
    SPI_RBIT(o, 15); \
    SPI_RBIT(o, 14); \
    SPI_RBIT(o, 13); \
    SPI_RBIT(o, 12); \
    SPI_RBIT(o, 11); \
    SPI_RBIT(o, 10); \
    SPI_RBIT(o, 9); \
    SPI_RBIT(o, 8); \
    SPI_RBIT(o, 23); \
    SPI_RBIT(o, 22); \
    SPI_RBIT(o, 21); \
    SPI_RBIT(o, 20); \
    SPI_RBIT(o, 19); \
    SPI_RBIT(o, 18); \
    SPI_RBIT(o, 17); \
    SPI_RBIT(o, 16); \
    SPI_RBIT(o, 31); \
    SPI_RBIT(o, 30); \
    SPI_RBIT(o, 29); \
    SPI_RBIT(o, 28); \
    SPI_RBIT(o, 27); \
    SPI_RBIT(o, 26); \
    SPI_RBIT(o, 25); \
    SPI_RBIT(o, 24);

static int card_avail;              /*!< Card is available. */
static int card_changed;            /*!< Card has changed. */

/*!
 * \brief Initialize MMC hardware interface.
 *
 * This function is automatically executed during during device
 * registration via NutRegisterDevice().
 *
 * \param dev Identifies the device to initialize.
 */
int SpiIfcInit(void)
{
    outr(PMC_PCER, _BV(PIOA_ID));

    /* Enable all SPI ports, if a port enable register is defined. */
    outr(SPI_PE_REG, _BV(SPI_CLK_BIT) | _BV(SPI_MOSI_BIT) | _BV(SPI_CS_BIT) | _BV(SPI_MISO_BIT));

    /* Enable CLK, MOSI and CS output pins. */
    outr(SPI_OE_REG, _BV(SPI_CLK_BIT) | _BV(SPI_MOSI_BIT) | _BV(SPI_CS_BIT));
    /* Enable MISO input pin. */
    outr(SPI_OD_REG, _BV(SPI_MISO_BIT));

    /* Set all outputs high. */
    outr(SPI_SOD_REG, _BV(SPI_CLK_BIT) | _BV(SPI_MOSI_BIT) | _BV(SPI_CS_BIT));

    card_avail = 1;
    card_changed = 0;

    return 0;
}

/*
 * Re-initializes ports bits in case they are shared.
 */
void SpiSelect(int on)
{
    /* Enable MISO input pin. */
    outr(SPI_OD_REG, _BV(SPI_MISO_BIT));
    /* Set CLK low, MOSI high (SPI mode 0). */
    outr(SPI_COD_REG, _BV(SPI_CLK_BIT));
    outr(SPI_SOD_REG, _BV(SPI_MOSI_BIT));
    /* MMC select is low active. */
    if (on == 1) {
        outr(SPI_COD_REG, _BV(SPI_CS_BIT));
    } else if (on == 0) {
        outr(SPI_SOD_REG, _BV(SPI_CS_BIT));
    }
    /* Enable CLK, MOSI and CS output pins. */
    outr(SPI_OE_REG, _BV(SPI_CLK_BIT) | _BV(SPI_MOSI_BIT) | _BV(SPI_CS_BIT));
}

/*!
 * \brief Initialize the card in slot 0.
 *
 * Called by the MMC driver during card reset. The card change
 * detection flag will be cleared.
 *
 * \return 0 on success, -1 if no card is detected.
 */
int SbiMmCard0Init(void)
{
    card_changed = 0;
    if (card_avail) {
        return 0;
    }
    printf("[NoCard]");
    return -1;
}

static unsigned char SpiReadByte(void)
{
    unsigned int rc = 0;

    SPI_R8BIT(rc);

    return (unsigned char)rc;
}

#define SPI_WBIT(i, n) \
    if ((i) & _BV(n)) { \
        outr(SPI_SOD_REG, _BV(SPI_MOSI_BIT)); \
    } else { \
        outr(SPI_COD_REG, _BV(SPI_MOSI_BIT)); \
    } \
    outr(SPI_SOD_REG, _BV(SPI_CLK_BIT)); \
    outr(SPI_COD_REG, _BV(SPI_CLK_BIT));


static void SpiWriteByte(unsigned int val)
{
    SPI_WBIT(val, 7);
    SPI_WBIT(val, 6);
    SPI_WBIT(val, 5);
    SPI_WBIT(val, 4);
    SPI_WBIT(val, 3);
    SPI_WBIT(val, 2);
    SPI_WBIT(val, 1);
    SPI_WBIT(val, 0);
}

/*!
 * \brief Send command to multimedia card.
 *
 * \param ifc   Specifies the hardware interface.
 * \param cmd   Command code. See MMCMD_ macros.
 * \param param Optional command parameter.
 */
static void MmCardTxCmd(unsigned char cmd, unsigned long param)
{
    unsigned int tmo = 8192;
    unsigned char ch;

    /* Enable card select. */
    SpiSelect(1);

    /*
     * Repeat sending nothing until we receive nothing. Actually
     * it should be sufficient to send a single 0xFF value, but
     * running a loop seems to fix certain kind of sync problems.
     */
    while ((ch = SpiReadByte()) != 0xFF) {
        if (--tmo == 0) {
            printf("[MMCmd%u Timeout %02X]\n", cmd, ch);
            break;
        }
        if (tmo < 256) {
            printf("<R>");
            Delay(10);
        }
    }

    /* Send command and parameter. */
    SpiWriteByte(MMCMD_HOST | cmd);
    SpiWriteByte((unsigned char) (param >> 24));
    SpiWriteByte((unsigned char) (param >> 16));
    SpiWriteByte((unsigned char) (param >> 8));
    SpiWriteByte((unsigned char) param);
    /*
     * We are running with CRC disabled. However, the reset command must
     * be send with a valid CRC. Fortunately this command is sent with a
     * fixed parameter value of zero, which results in a fixed CRC value
     */
    SpiWriteByte(MMCMD_RESET_CRC);
}

/*!
 * \brief Receive an R1 response token from the card.
 *
 * In SPI mode, the card sends an R1 response after every command except
 * after the SEND_STATUS and the READ_OCR commands.
 *
 * \param ifc Specifies the hardware interface.
 *
 * \return R1 response token or 0xFF on timeout.
 */
unsigned char MmCardRxR1(void)
{
    unsigned char rc;
    int i;

    for (i = 0; i < 255; i++) {
        if ((rc = SpiReadByte()) != 0xFF) {
            break;
        }
    }
    return rc;
}

/*!
 * \brief Receive an R2 response token from the card.
 *
 * In SPI mode the card sends this token in response to the SEND_STATUS
 * command.
 *
 * \param ifc Specifies the hardware interface.
 *
 * \return R2 response token or 0xFFFF on timeout.
 */
unsigned short MmCardRxR2(void)
{
    unsigned short rc;

    rc = MmCardRxR1();
    rc <<= 8;
    rc += SpiReadByte();

    return rc;
}

/*!
 * \brief Receive an R3 response token from the card.
 *
 * In SPI mode the card sends this token in response to the READ_OCR
 * command.
 *
 * \param ifc Specifies the hardware interface.
 * \param ocr Points to a buffer which receives the OCR register contents.
 *
 * \return R1 response token or 0xFF on timeout.
 */
unsigned char MmCardRxR3(unsigned long * ocr)
{
    unsigned char rc;
    int i;

    /* The first byte is equal to the R1 response. */
    rc = MmCardRxR1();
    /* Receive the operating condition. */
    for (i = 0; i < 4; i++) {
        *ocr <<= 8;
        *ocr |= SpiReadByte();
    }
    return rc;
}

/*!
 * \brief Configure card for SPI mode.
 *
 * \param ifc Specifies the hardware interface.
 *
 * \return 0 on success, -1 otherwise.
 */
static int MmCardReset(void)
{
    int i;
    unsigned char rsp;

    /*
     * Initialize the low level card interface.
     */
    if (SbiMmCard0Init()) {
        return -1;
    }

    /*
     * 80 bits of ones with deactivated chip select will put the card 
     * in SPI mode.
     */
    SpiSelect(0);
    for (i = 0; i < 10; i++) {
        SpiReadByte();
    }

    /*
     * Switch to idle state and wait until initialization is running
     * or idle state is reached.
     */
    for (i = 0; i < 255; i++) {
        MmCardTxCmd(MMCMD_GO_IDLE_STATE, 0);
        rsp = MmCardRxR1();
        SpiSelect(0);
        if (rsp == MMR1_IDLE_STATE || rsp == MMR1_NOT_IDLE) {
            return 0;
        }
    }
    return -1;
}

/*!
 * \brief Initialize the multimedia card.
 *
 * This routine will put a newly powered up card into SPI mode.
 * It is called by MmCardMount().
 *
 * \param ifc Specifies the hardware interface.
 *
 * \return 0 on success, -1 otherwise.
 */
int MmCardInit(void)
{
    int i;
    unsigned char rsp;

    if (SpiIfcInit()) {
        return -1;
    }

    /*
     * Try to switch to SPI mode. Looks like a retry helps to fix
     * certain synchronization problems.
     */
    if (MmCardReset()) {
        if (MmCardReset()) {
            puts("CardReset failed");
            return -1;
        }
    }

    /*
     * Wait for a really long time until card is initialized
     * and enters idle state.
     */
    for (i = 0; i < 512; i++) {
        /*
         * In SPI mode SEND_OP_COND is a dummy, used to poll the card
         * for initialization finished. Thus, there are no parameters
         * and no operation condition data is sent back.
         */
        MmCardTxCmd(MMCMD_SEND_OP_COND, 0);
        rsp = MmCardRxR1();
        SpiSelect(0);
        if (rsp == MMR1_IDLE_STATE) {
            return 0;
        }
        if (i > 128) {
            Delay(1);
        }
    }
    puts("CardInit failed");
    return -1;
}

/*!
 * \brief Retrieve contents of a card register.
 *
 * \param ifc Specifies the hardware interface.
 * \param cmd This command is sent to the card to retrieve the contents
 *            of a specific register.
 * \param rbp Pointer to the buffer that receives the register contents.
 * \param siz Size of the specified register.
 *
 * \return 0 on success, -1 otherwise.
 */
static int MmCardGetReg(unsigned char cmd, unsigned char * rbp, int siz)
{
    int rc = -1;
    int retries = 512;
    int i;

    while (retries--) {
        /* Send the command to the card. This will select the card. */
        MmCardTxCmd(cmd, 0);
        /* Wait for OK response from the card. */
        if (MmCardRxR1() == 0x00) {
            /* Wait for data from the card. */
            if (MmCardRxR1() == 0xFE) {
                for (i = 0; i < siz; i++) {
                    *rbp++ = SpiReadByte();
                }
                /* Ignore the CRC. */
                SpiReadByte();
                SpiReadByte();
                /* De-activate card selection. */
                SpiSelect(0);
                /* Return a positive result. */
                rc = 0;
                break;
            }
        }
        /* De-activate card selection. */
        SpiSelect(0);
    }
    return rc;
}

int QueryCard(void)
{
    int rc;
    MMC_CID cid;

    rc = MmCardGetReg(MMCMD_SEND_CID, (unsigned char *)&cid, sizeof(MMC_CID));
    if (rc == 0) {
        printf("Manufacturer : %u\n", cid.mmcid_mid);
        printf("OEM ID       : %u\n", cid.mmcid_oid);
        printf("Product name : %.6s\n", cid.mmcid_pnm);
        printf("Product rev. : %u\n", cid.mmcid_rev);
        printf("Serial number: %lu\n", cid.mmcid_psn);
        printf("Date Code    : %u\n", cid.mmcid_mdt);
    }
    return rc;
}
