/*
 * Copyright (C) 2008 by egnite GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Id$
 */
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "eirhwt.h"

#define BUF 16

#define	LADJUST		0x04    /* left adjustment */
#define	ALT		0x01    /* alternate form */
#define	ZEROPAD		0x10    /* zero (as opposed to blank) pad */
#define	LONGINT		0x08    /* long integer */

#define	PADSIZE	16
static char blanks[PADSIZE] = { ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ',
    ' ', ' '
};
static char zeroes[PADSIZE] = { '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
    '0', '0'
};

static void _putpad(char *padch, int count)
{
    while (count > PADSIZE) {
        DbgWrite(padch, PADSIZE);
        count -= PADSIZE;
    }
    if (count > 0)
        DbgWrite(padch, count);
}

int vprintf(const char *fmt, va_list ap)
{
    unsigned char ch;                  /* character from fmt */
    int n;                      /* handy integer (short term usage) */
    char *cp;                   /* handy char pointer (short term usage) */
    unsigned char flags;               /* flags as above */
    int rc;                     /* return value accumulator */
    int width;                  /* width from format (%8d), or 0 */
    int prec;                   /* precision from format (%.3d), or -1 */
    int dprec;                  /* a copy of prec if [diouxX], 0 otherwise */
    int realsz;                 /* field size expanded by dprec, sign, etc */
    unsigned char sign;                /* sign prefix (' ', '+', '-', or \0) */
    unsigned long ulval;               /* integer arguments %[diouxX] */
    int size;                   /* size of converted field or string */
    char *xdigs;                /* digits for [xX] conversion */
    char buf[BUF];              /* space for %c, %[diouxX], %[eEfgG] */

    rc = 0;

    for (;;) {

        /*
         * Print format string until next percent sign.
         */
        for (cp = (char *) fmt; (ch = *fmt) != 0 && ch != '%'; fmt++);
        if ((n = fmt - cp) != 0) {
            DbgWrite(cp, n);
            rc += n;
        }
        if (ch == 0)
            break;
        fmt++;

        /*
         * Process modifiers.
         */
        flags = 0;
        sign = 0;
        width = 0;
        dprec = 0;
        prec = -1;
        for (;;) {
            ch = *fmt++;
            if (ch == ' ') {
                if (!sign)
                    sign = ' ';
            } else if (ch == '+')
                sign = '+';
            else if (ch == '-')
                flags |= LADJUST;
            else if (ch == '#')
                flags |= ALT;
            else if (ch == '0')
                flags |= ZEROPAD;
            else if (ch == 'l')
                flags |= LONGINT;
            else if (ch == 'z') {
                if (sizeof(size_t) > sizeof(int)) {
                    flags |= LONGINT;
                }
            }
            else if (ch == '*') {
                width = va_arg(ap, int);
                if (width < 0) {
                    flags |= LADJUST;
                    width = -width;
                }
            } else if (ch == '.') {
                if (*fmt == '*') {
                    fmt++;
                    prec = va_arg(ap, int);
                } else {
                    prec = 0;
                    while (*fmt >= '0' && *fmt <= '9')
                        prec = 10 * prec + (*fmt++ - '0');
                }
                if (prec < 0)
                    prec = -1;
            } else if (ch >= '1' && ch <= '9') {
                width = ch - '0';
                while (*fmt >= '0' && *fmt <= '9')
                    width = 10 * width + (*fmt++ - '0');
            } else
                break;
        }

        /*
         * Process type field.
         */
        switch (ch) {
        case 'c':
            *(cp = buf) = va_arg(ap, int);
            size = 1;
            sign = 0;
            break;
            
        case 's':
            cp = va_arg(ap, char *);

            if (cp == 0)
                cp = "(null)";
            if (prec >= 0) {
                char *p = memchr(cp, 0, (size_t) prec);

                if (p) {
                    size = p - cp;
                    if (size > prec)
                        size = prec;
                } else
                    size = prec;
            } else
                size = strlen(cp);
            sign = 0;
            break;

        case 'u':
            sign = 0;
        case 'd':
        case 'i':
            /* Thanks to Ralph Mason for fixing the unsigned int bug. */
            if (flags & LONGINT)
                ulval = va_arg(ap, unsigned long);
            else if (ch == 'u')
                ulval = va_arg(ap, unsigned int);
            else
                ulval = va_arg(ap, int);
            if (ch != 'u' && (long) ulval < 0) {
                ulval = (unsigned long) (-((long) ulval));
                sign = '-';
            }
            if ((dprec = prec) >= 0)
                flags &= ~ZEROPAD;
            cp = buf + BUF;
            if (ulval || prec) {
                if (ulval < 10)
                    *--cp = (char) ulval + '0';
                else
                    do {
                        *--cp = (char) (ulval % 10) + '0';
                        ulval /= 10;
                    } while (ulval);
            }
            size = buf + BUF - cp;
            break;

        case 'o':
            ulval = (flags & LONGINT) ? va_arg(ap, unsigned long) : va_arg(ap, unsigned int);
            sign = 0;
            if ((dprec = prec) >= 0)
                flags &= ~ZEROPAD;
            cp = buf + BUF;
            if (ulval || prec) {
                do {
                    *--cp = (char) (ulval & 7) + '0';
                    ulval >>= 3;
                } while (ulval);
                if ((flags & ALT) != 0 && *cp != '0')
                    *--cp = '0';
            }
            size = buf + BUF - cp;
            break;

        case 'p':
        case 'X':
        case 'x':
            if (ch == 'p') {
                ulval = (unsigned __PTRDIFF_TYPE__) va_arg(ap, void *);
                flags |= ALT;
                ch = 'x';
            } else
                ulval = (flags & LONGINT) ? va_arg(ap, unsigned long) : (unsigned long)
                    va_arg(ap, unsigned int);

            sign = 0;
            if ((dprec = prec) >= 0)
                flags &= ~ZEROPAD;

            if (ch == 'X')
                xdigs = "0123456789ABCDEF";
            else
                xdigs = "0123456789abcdef";

            cp = buf + BUF;
            do {
                *--cp = xdigs[ulval & 0x0f];
                ulval >>= 4;
            } while (ulval);
            if (flags & ALT) {
                *--cp = ch;
                *--cp = '0';
            }
            size = buf + BUF - cp;
            break;

        case 'g':
        case 'G':
        case 'e':
        case 'E':
        case 'f':
            (void) va_arg(ap, long);

        default:
            if (ch == 0)
                return rc;
            cp = buf;
            *cp = ch;
            size = 1;
            sign = '\0';
            break;
        }                       /* switch */

        /*
         * Output.
         */
        realsz = dprec > size ? dprec : size;
        if (sign)
            realsz++;

        if ((flags & (LADJUST | ZEROPAD)) == 0)
            _putpad(blanks, width - realsz);

        if (sign)
            DbgWrite(&sign, 1);

        if ((flags & (LADJUST | ZEROPAD)) == ZEROPAD)
            _putpad(zeroes, width - realsz);

        _putpad(zeroes, dprec - size);

        if (size)		/* DF 12/16/03 - zero length is "flush" in NutTcpDeviceWrite() */
            DbgWrite(cp, size);

        if (flags & LADJUST)
            _putpad(blanks, width - realsz);

        if (width >= realsz)
            rc += width;
        else
            rc += realsz;
    }
    return rc;
}

/*!
 * \brief Print formatted data to the standard output stream.
 *
 * \param fmt Format string containing conversion specifications.
 *
 * \return The number of characters written or a negative value to
 *         indicate an error.
 */
int printf(const char *fmt, ...)
{
    int rc;
    va_list ap;

    va_start(ap, fmt);
    rc = vprintf(fmt, ap);
    va_end(ap);

    return rc;
}

int puts(const char *string)
{
    DbgWrite(string, strlen(string));
    putchar('\n');

    return 0;
}

int putchar(int c)
{
    char ch = (char) c;

    DbgWrite(&ch, 1);

    return c;
}
