/*
 * Copyright (C) 2008 by egnite GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Id$
 */
#include <stdlib.h>
#include <string.h>

#include "eirhwt.h"

/* DataFlash at PA11_SPI0_NPCS0_A */
#define VS_DREQ_BIT     PA30_IRQ1_A         /* PA30 */
#define VS_XCS_BIT      PA31_SPI0_NPCS1_A   /* PA31 */
#define VS_XDCS_BIT     PB30_SPI0_NPCS2_A   /* PB30 */
#define VS_XRESET_BIT   31                  /* PB31 */

/*
 * Instruction opcodes.
 */
#define VS_OPCODE_READ  3
#define VS_OPCODE_WRITE 2

#define VS_MODE_REG         0
#define VS_STATUS_REG       1

#define VS10XX_DREQ_TST()    ((inr(PIOA_PDSR) & _BV(VS_DREQ_BIT)) == _BV(VS_DREQ_BIT))

static unsigned char SciByte(unsigned char val)
{
#if 0
    putchar('[');
    if (val != 0xFF) {
        printf("s%02X", val);
    }
#endif

    /* Transmission is started by writing the transmit data. */
    outr(SPI0_TDR, val);
    /* Wait for receiver data register full. */
    while((inr(SPI0_SR) & SPI_RDRF) == 0);
    /* Read data. */
    val = (unsigned char)inr(SPI0_RDR);

#if 0
    if (val != 0xFF) {
        printf("r%02X", val);
    }
    putchar(']');
#endif
    return val;
}

static int VsWaitReady(void)
{
    int tmo;

    for (tmo = 0; tmo < 5000; tmo++) {
        if (VS10XX_DREQ_TST()) {
            return 0;
        }
    }
    return -1;
}

void SdiDeselect(void)
{
    outr(PIOB_SODR, _BV(VS_XDCS_BIT));
}

void SdiSelect(void)
{
    outr(PIOB_CODR, _BV(VS_XDCS_BIT));
}

void SciDeselect(void)
{
    outr(PIOA_SODR, _BV(VS_XCS_BIT));
}

void SciSelect(void)
{
    outr(PIOA_CODR, _BV(VS_XCS_BIT));
}

static void VsSciSelect(unsigned int on)
{
    if (on) {
        /* Disable data interface. */
        SdiDeselect();
        /* Enable cmd interface. */
        SciSelect();
    }
    else {
        /* Disable cmd interface. */
        SciDeselect();
        /* Enable data interface. */
        SdiSelect();
    }
}

static unsigned short VsRegRead(unsigned int reg)
{
    unsigned short data;

    /* Select command channel. */
    VsWaitReady();
    VsSciSelect(1);

    SciByte(VS_OPCODE_READ);
    SciByte((unsigned char) reg);
    data = (unsigned short)SciByte(0) << 8;
    data |= SciByte(0);

    /* Select data channel. */
    VsSciSelect(0);

    return data;
}

static void SpiInit(void)
{
    /* Enable peripherals. */
    outr(PIOA_ASR, inr(PIOA_ASR) 
        | _BV(PA12_SPI0_MISO_A) 
        | _BV(PA13_SPI0_MOSI_A)
        | _BV(PA14_SPI0_SPCK_A));
        //| _BV(PA31_SPI0_NPCS1_A);
        //| _BV(PA30_IRQ1_A);

    /* Disable PIO lines used for SPI. */
    outr(PIOA_PDR, _BV(PA12_SPI0_MISO_A)
        | _BV(PA13_SPI0_MOSI_A)
        | _BV(PA14_SPI0_SPCK_A));
        //| _BV(PA31_SPI0_NPCS1_A);
        //| _BV(PA30_IRQ1_A);

    /* Chip selects are manually controlled. */
    /* Enable XCS control. */
    outr(PIOA_PER, _BV(VS_XCS_BIT));
    outr(PIOA_SODR, _BV(VS_XCS_BIT));
    outr(PIOA_OER, _BV(VS_XCS_BIT));

    /* Enable XDCS control. */
    outr(PIOB_PER, _BV(VS_XDCS_BIT));
    outr(PIOB_SODR, _BV(VS_XDCS_BIT));
    outr(PIOB_OER, _BV(VS_XDCS_BIT));

    /* Release reset line. */
    outr(PIOB_PER, _BV(VS_XRESET_BIT));
    outr(PIOB_SODR, _BV(VS_XRESET_BIT));
    outr(PIOB_OER, _BV(VS_XRESET_BIT));

    /* Enable SPI clock. */
    outr(PMC_PCER, _BV(SPI0_ID));

    /* SPI enable and reset. */
    outr(SPI0_CR, SPI_SPIEN | SPI_SWRST);
    outr(SPI0_CR, SPI_SPIEN);

    /* Set SPI to master mode, fixed peripheral, fault detection disabled. */
    outr(SPI0_MR, (1 << SPI_PCS_LSB) | SPI_MODFDIS | SPI_MSTR);

    /* Data changes during clock low and will be sampled on rising edges. */
    outr(SPI0_CSR1, (50 << SPI_SCBR_LSB) | SPI_NCPHA);
    outr(SPI0_CSR2, (50 << SPI_SCBR_LSB) | SPI_NCPHA);
}

int Vs10x3Check(void)
{
    unsigned short mode_reg;
    unsigned short stat_reg;

    printf("Detect VS10X3...");
    SpiInit();
    mode_reg = VsRegRead(VS_MODE_REG);
    if (mode_reg != 0x4800) {
        puts("Error");
        printf("VS Mode = %04X\n", mode_reg);
        return -1;
    }
    stat_reg = VsRegRead(VS_STATUS_REG);
    if ((stat_reg & 0xfff0) != 0x0040) {
        puts("Error");
        printf("VS Status = %04X\n", stat_reg);
        return -1;
    }
    puts("OK");

    return 0;
}
